#ifndef __ASM_PAGE_H__
#define __ASM_PAGE_H__


#define PAGE_SHIFT 12
#define PAGE_SIZE  ( 1 << PAGE_SHIFT )
#define PAGE_MASK  ( ~ ( PAGE_SIZE - 1 ) )

#define PAGE_SHIFT_2MB 21

#define PFN_UP(x)	(((x) + PAGE_SIZE - 1) >> PAGE_SHIFT)
#define PFN_DOWN(x)	((x) >> PAGE_SHIFT)
#define PFN_PHYS(x)	((x) << PAGE_SHIFT)

#define PAGE_UP(p)    ( ( (p) + ( PAGE_SIZE - 1 ) ) & PAGE_MASK )
#define PAGE_DOWN(p)  ( (p) & PAGE_MASK )



/* Page-Translation-Table Entry Fields
   [REF] vol.2, p. 168- */
#define _PTTEF_PRESENT   0
#define _PTTEF_RW        1 /* Read/Write */
#define _PTTEF_US        2 /* User/Supervisor */
#define _PTTEF_ACCESSED  5
#define _PTTEF_DIRTY     6
#define _PTTEF_PAGE_SIZE 7
#define _PTTEF_GLOBAL	 8
#define PTTEF_PRESENT    (1 << _PTTEF_PRESENT)
#define PTTEF_RW         (1 << _PTTEF_RW)
#define PTTEF_US         (1 << _PTTEF_US)
#define PTTEF_ACCESSED   (1 << _PTTEF_ACCESSED)
#define PTTEF_DIRTY      (1 << _PTTEF_DIRTY)
#define PTTEF_PAGE_SIZE  (1 << _PTTEF_PAGE_SIZE)
#define PTTEF_GLOBAL     (1 << _PTTEF_GLOBAL)

/* [REF] vol. 2, p. 59 */
#define _X86_CR4_PSE 4 /* Page Size Extensions */
#define _X86_CR4_PAE 5 /* Physical-Address Extension */
#define _X86_CR4_PGE 7 /* Page-Global Enable */
#define X86_CR4_PSE  (4 << _X86_CR4_PSE) 
#define X86_CR4_PAE  (5 << _X86_CR4_PAE) 
#define X86_CR4_PGE  (7 << _X86_CR4_PGE) 


/* [REF] vol. 2, p. 53 */
#define _X86_CR0_PE  0 /* Protection enabled */
#define _X86_CR0_MP  1 /* Monitor coprocessor */
#define _X86_CR0_ET  4 /* Extension type */
#define _X86_CR0_NE  5 /* Numeric error */
#define _X86_CR0_WP 16 /* Write protected */
#define _X86_CR0_AM 18 /* Aligned mask */
#define _X86_CR0_PG 31 /* Paging */
#define X86_CR0_PE  (1 << _X86_CR0_PE) 
#define X86_CR0_MP  (1 << _X86_CR0_MP) 
#define X86_CR0_ET  (1 << _X86_CR0_ET) 
#define X86_CR0_NE  (1 << _X86_CR0_NE) 
#define X86_CR0_WP  (1 << _X86_CR0_WP) 
#define X86_CR0_AM  (1 << _X86_CR0_AM) 
#define X86_CR0_PG  (1 << _X86_CR0_PG) 


#ifdef __ASSEMBLY__

#define VMM_OFFSET	0xFFFF830000000000
#define PHYS(va)	((va) - VMM_OFFSET)

#else /* ! __ASSEMBLY__ */

#define VMM_OFFSET	0xFFFF830000000000UL
#define PHYS(va)	((unsigned long)(va) - VMM_OFFSET)
#define VIRT(pa)	((void *)((unsigned long)(pa) + VMM_OFFSET))

#endif /* __ASSEMBLY__ */



#ifndef __ASSEMBLY__

unsigned long pml4_table_create ( void );
extern void mmap ( unsigned long pml4_table_base_vaddr, unsigned long vaddr, unsigned long paddr );

#endif /* ! __ASSEMBLY__ */



#endif /* __ASM_PAGE_H__ */
