#include "types.h"
#include "bitops.h"
#include "string.h"
#include "printf.h"
#include "failure.h"
#include "page.h"
#include "msr.h"
#include "processor.h"
#include "cpufeature.h"
#include "svm.h"
#include "alloc.h"


/* Host save area */
static void *host_save_area;


/* ҡפ˳ݤɬפ롩 */
void *
alloc_host_save_area ( void )
{
	void *hsa;

	unsigned long n  = alloc_pages ( 1, 1 );
	hsa = ( void * ) VIRT ( n << PAGE_SHIFT );
	
	if ( hsa ) {
		memset ( hsa, 0, PAGE_SIZE );
	}
	
	return hsa;
}

void __init
enable_svm ( struct cpuinfo_x86 *c )
{
 	/* Xen does not fill x86_capability words except 0. */
	{
		u32 ecx = cpuid_ecx ( 0x80000001 );
		c->x86_capability[5] = ecx;
	}
    
	if ( ! ( test_bit ( X86_FEATURE_SVME, &c->x86_capability ) ) ) {
		fatal_failure ( "No svm featurel\n" );
		return;
	}
   
	{
		u32 eax, edx;
		rdmsr ( MSR_EFER, eax, edx );
		eax |= EFER_SVME;
		wrmsr ( MSR_EFER, eax, edx );
	}

	printf ( "AMD SVM Extension is enabled.\n" );

	/* Initialize the HSA */
	{
		u64 phys_hsa;
		u32 phys_hsa_lo, phys_hsa_hi;   

		host_save_area = alloc_host_save_area ( );
		phys_hsa = ( u64 ) PHYS ( host_save_area ); 
		phys_hsa_lo = ( u32 ) phys_hsa;
		phys_hsa_hi = ( u32 ) (phys_hsa >> 32);    
		wrmsr ( MSR_K8_VM_HSAVE_PA, phys_hsa_lo, phys_hsa_hi ); 
	}
}

#if 0

/*****************************************************************************************************/

void 
svm_final_setup_guest ( struct vcpu *v )
{
	struct domain *d = v->domain;
	struct vcpu *vc;
	
	/* Initialize monitor page table */
	for_each_vcpu(d, vc)
		vc->arch.monitor_table = mk_pagetable(0);
	
	/* 
	 * Required to do this once per domain
	 * TODO: add a seperate function to do these.
	 */
	memset ( &d->shared_info->evtchn_mask[0], 0xff, sizeof ( d->shared_info->evtchn_mask ) );
	
	/* 
	 * Put the domain in shadow mode even though we're going to be using
	 * the shared 1:1 page table initially. It shouldn't hurt 
	 */
	shadow_mode_enable ( d, ( SHM_enable | SHM_refcounts | SHM_translate | SHM_external | SHM_wr_pt_pte ) );
}

/*****************************************************************************************************/



/* Make sure that xen intercepts any FP accesses from current */
static void 
svm_stts ( struct vcpu *v ) 
{
	/* FPU state already dirty? Then no need to setup_fpu() lazily. */
	if ( test_bit(_VCPUF_fpu_dirtied, &v->vcpu_flags) ) {
		return;
	}

	/*
	 * If the guest does not have TS enabled then we must cause and handle an 
	 * exception on first use of the FPU. If the guest *does* have TS enabled 
	 * then this is not necessary: no FPU activity can occur until the guest 
	 * clears CR0.TS, and we will initialise the FPU when that happens.
	 */
	if ( ! ( v->arch.hvm_svm.cpu_shadow_cr0 & X86_CR0_TS ) ) {
		v->arch.hvm_svm.vmcb->exception_intercepts |= EXCEPTION_BITMAP_NM;
		vmcb->cr0 |= X86_CR0_TS;
	}
}

/*****************************************************************************************************/

void
hvm_setup_platform ( struct domain* d )
{
	struct hvm_domain *platform;

	if ( ! hvm_guest ( current ) || ( current->vcpu_id != 0 ) ) {
		return;
	}

	shadow_direct_map_init(d);

	hvm_map_io_shared_page(d);
	hvm_get_info(d);

	platform = &d->arch.hvm_domain;
	pic_init(&platform->vpic, pic_irq_request, &platform->interrupt_request);
	register_pic_io_hook();

	if ( hvm_apic_support(d) ) {
		spin_lock_init(&d->arch.hvm_domain.round_robin_lock);
		hvm_vioapic_init(d);
	}
}

static int 
svm_paging_enabled ( struct vcpu *v )
{
	unsigned long cr0;

	cr0 = v->arch.hvm_svm.cpu_shadow_cr0;

	return ( cr0 & X86_CR0_PE ) && ( cr0 & X86_CR0_PG );
}

void 
svm_do_launch ( struct vcpu *v )
{
	struct vmcb_struct *vmcb = v->arch.hvm_svm.vmcb;
	int core = smp_processor_id();
	
	/* Update CR3, GDT, LDT, TR */
	svm_stts(v);
	
	/* current core is the one we intend to perform the VMRUN on */
//	v->arch.hvm_svm.launch_core = v->arch.hvm_svm.asid_core = core;
//	clear_bit ( ARCH_SVM_VMCB_ASSIGN_ASID, &v->arch.hvm_svm.flags );
//	asidpool_assign_next( vmcb, 0, core, core );

	hvm_setup_platform ( v->domain );
	
//	evtchn_bind_vcpu ( iopacket_port(v), v->vcpu_id );
//	clear_bit ( iopacket_port(v), &v->domain->shared_info->evtchn_mask[0]);
//	
//	if (hvm_apic_support(v->domain)) {
//		vlapic_init(v);
//	}

	init_timer(&v->arch.hvm_svm.hlt_timer,  hlt_timer_fn, v, v->processor);
	
	vmcb->ldtr.sel = 0;
	vmcb->ldtr.base = 0;
	vmcb->ldtr.limit = 0;
	vmcb->ldtr.attributes.bytes = 0;
	
	vmcb->efer = EFER_SVME; /* Make sure VMRUN won't return with -1 */
	
	vmcb->cr3 = pagetable_get_paddr ( svm_paging_enabled(v) ? v->arch.guest_table : ov->domain->arch.phys_table );

//	v->arch.hvm_svm.injecting_event  = 0;
	v->arch.hvm_svm.saved_irq_vector = -1;
	
	svm_set_guest_time ( v, 0 );
	
	vmcb->tlb_control = 1;
}

static void
arch_svm_do_launch(struct vcpu *v) 
{
	cpu_user_regs_t *regs = &current->arch.guest_context.user_regs;
	
	construct_vmcb ( &v->arch.hvm_svm, regs );
	
	svm_do_launch ( v );
	
	reset_stack_and_jump ( svm_asm_do_launch ); 
}

/*********************************************************************************************/

static void 
svm_freeze_time ( struct vcpu *v )
{
	struct hvm_virpit *vpit = &v->domain->arch.hvm_domain.vpit;
	
	if ( vpit->first_injected && !v->domain->arch.hvm_domain.guest_time ) {
		v->domain->arch.hvm_domain.guest_time = svm_get_guest_time(v);
		stop_timer(&(vpit->pit_timer));
	}
}

static void 
svm_ctxt_switch_from ( struct vcpu *v )
{
	svm_freeze_time(v);
}

/*********************************************************************************************/

static void 
svm_ctxt_switch_to ( struct vcpu *v )
{
	/* 
	 * This is required, because VMRUN does consistency check
	 * and some of the DOM0 selectors are pointing to 
	 * invalid GDT locations, and cause AMD processors
	 * to shutdown.
	 */
	set_segment_register(ds, 0);
	set_segment_register(es, 0);
	set_segment_register(ss, 0);
}

/*********************************************************************************************/

static void
svm_do_resume ( struct vcpu *v ) 
{
	struct domain *d = v->domain;
	struct hvm_virpit *vpit = &d->arch.hvm_domain.vpit;

	svm_stts(v);

	/* make sure the HSA is set for the current core */
	set_hsa_to_guest( &v->arch.hvm_svm );
    
	/* pick up the elapsed PIT ticks and re-enable pit_timer */
	if ( vpit->first_injected ) {
		if ( v->domain->arch.hvm_domain.guest_time ) {
			svm_set_guest_time(v, v->domain->arch.hvm_domain.guest_time);
			v->domain->arch.hvm_domain.guest_time = 0;
		}
		pickup_deactive_ticks(vpit);
	}

	if ( test_bit(iopacket_port(v), &d->shared_info->evtchn_pending[0]) || test_bit(ARCH_HVM_IO_WAIT, &v->arch.hvm_vcpu.ioflags) ) {
		hvm_wait_io();
	}
}

void
arch_svm_do_resume ( struct vcpu *v ) 
{
	svm_do_resume ( v );
	reset_stack_and_jump( svm_asm_do_resume );
}

#endif /* 0 */

