#include "string.h"
#include "printf.h"
#include "page.h"
#include "multiboot.h"
#include "elf.h"
#include "vm.h"
#include "alloc.h"


static u32
init_vm_mbi ( unsigned long vm_pmem_start )
{
	enum { INSTALL_PADDR = 0x2d0e0UL }; /* < 1 MB */
	struct multiboot_info *mbi = ( struct multiboot_info *) ( vm_pmem_start + INSTALL_PADDR );

	printf ( "Initializing multiboot info.\n" );

	return PHYS ( mbi );
}

void
vm_create ( struct vm *vm, unsigned long guest_image_start, unsigned long guest_image_size, unsigned long vm_pmem_size )
{
	struct vmcb_struct *vmcb = &vm->vmcb;

	printf ( "Creating virtual machine.\n" ); 

	/* ۥޥѤʪݤ */
	const unsigned long vm_pmem_pfn = alloc_pages ( vm_pmem_size >> PAGE_SHIFT, 1 );
	unsigned long vm_pmem_start     = ( unsigned long ) VIRT ( vm_pmem_pfn << PAGE_SHIFT );

	/* Copy the OS image to the specified region by interpreting the ELF format.  */
	vmcb->rip = load_elf_image ( guest_image_start, guest_image_size, vm_pmem_start );

#if 0
	/* Setup multiboot info.  */
	vmcb->rbx = init_vm_mbi ( vm_pmem_start );
#endif

	/* [???] ٥Υڡ󥰤ˤФ褤Τ */
	vmcb->h_cr3 = pml4_table_create ( );
	int i;
	for ( i = 0; i < ( vm_pmem_size >> PAGE_SHIFT ); i++ ) {
		const unsigned long pml4     = ( unsigned long ) VIRT ( vmcb->h_cr3 );
		const unsigned long vm_paddr = i << PAGE_SHIFT;
		const unsigned long pm_paddr = ( vm_pmem_pfn + i ) << PAGE_SHIFT;
		mmap ( pml4, vm_paddr, pm_paddr );
	}

	/* Enable nested paging (See AMD64 manual Vol. 3, p. 488) */
	vmcb->np_enable = 1; 
}

void
vm_boot ( struct vm *vm )
{
	printf ( "Booting guest operating system.\n" ); 

	u32 rax = PHYS ( &vm->vmcb );
	__asm__ __volatile__ ( "movl %%eax, %0" : "=r" (rax) :  );

	__asm__ __volatile__ ( "jmp svm_launch" : : );
}
